
package pi7;

import java.util.Map;
import java.util.HashMap;
import gnu.io.CommPort;
import gnu.io.CommPortIdentifier;
import gnu.io.SerialPort;

import java.io.FileDescriptor;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;




public class CommunicationController {

    public enum Register { INICIAR, CONTINUAR, SUSPENDER, X, Y, PROGRAM_LINE };
    public static int SUCCESS = 0;
    public static int LRC_ERROR = -1;
    public static int TIMEOUT_ERROR = -2;

    private Map<Register, Integer> addressMap = new HashMap<Register, Integer>();
    private String commPortName = null;
    private SerialPort serialPort = null;
    private InputStream inputSerialStream = null;
    private OutputStream outputSerialStream = null;

    private static int NUMBER_OF_RETRIES = 3;
    private static long FRAME_TIMEOUT = 60000;
    private String sync = new String();

    private enum FunctionCode { ReadRegister, WriteRegister, WriteFile };

    public CommunicationController() {
       // Ver "mapa de memoria" nos slides da aula
       // TO_DO: completar "mapa de memoria"
       addressMap.put( Register.X, 8); 
       addressMap.put( Register.Y, 9);
       addressMap.put( Register.PROGRAM_LINE, 10);
    } // constructor

    public void init(String commPort) throws Exception {
       this.commPortName = commPort;
       connect(commPort);
    } // setCommParameters

    public CommReturn writeRegister(int destination, Register register, int value)throws Exception {
         CommReturn ret = null;
        // TO_DO : completar
        synchronized(sync) {
        }
        return ret;
    } // writeRegister

    public CommReturn readRegister(int destination, Register register) throws Exception {
        CommReturn ret = null;
        synchronized(sync) {
          byte[] frame = buildReadRegisterFrame(destination, register);
          ret = sendRequestGetResponse(frame, FunctionCode.ReadRegister);
        }
        return ret;
    } // readRegister

    public CommReturn writeFile(int destination, byte[] data) throws Exception {
        // TO_DO : completar
        CommReturn ret = null;
        synchronized(sync) {
        }
        return ret;
    } // writeFile

    private CommReturn sendRequestGetResponse(byte[] frame, FunctionCode functionCode) throws Exception {
       CommReturn ret = new CommReturn();
       ret.success = false;
       ret.exceptionCode = TIMEOUT_ERROR; // timeout
       for (int retries = 0; retries < NUMBER_OF_RETRIES; retries++) {
           sendRequest(frame);
           ret = getResponse();
           if (ret.success) {
               if (checkLRC(ret)) {
                 unassemblyResponse(ret, functionCode);
                 break;                  
               } else {
                   ret.exceptionCode = LRC_ERROR;
               }
           } else {
               System.out.println("Timeout");
           }
       } // for retries
       return ret;
    } // sendRequestGetResponse

    private void sendRequest(byte[] frame) throws Exception {
        outputSerialStream.write(frame);
        outputSerialStream.flush();
    } // sendRequest

    private CommReturn getResponse() throws Exception {
        CommReturn ret = new CommReturn();
        byte[] buffer = new byte[1024];
        int idx = 0;
        Timer timer = new Timer(FRAME_TIMEOUT);
        timer.start();

        // hunt for ':'
        boolean found = false;
        while ( ( !found) && ( ! timer.timedOut())) {
            if ( inputSerialStream.available() > 0 ) {
                int chInt = inputSerialStream.read();
                System.out.println("got " + (char)chInt);
                if ( chInt != -1) {
                    if ( ':' == (byte)chInt ) {
                        buffer[0] = (byte)chInt;
                        found = true;
                        idx++;
                    }
                } else {
                    throw new Exception ("Port closed");
                }
            } else {
                Thread.sleep(10);
            }
        } // while !found and !timeout when hunting ':'

        // hunt for CR+LF
        if ( ! timer.timedOut() ) {
            found = false;
            while ( ( !found)  && ( !timer.timedOut())) {
               if ( inputSerialStream.available() > 0 ) {
                   int chInt = inputSerialStream.read();
                   if ( chInt != -1) {
                       System.out.println("got " + (char)chInt + " " + Integer.toHexString(chInt));
                       buffer[idx] = (byte) chInt;
                       if ( (buffer[idx-1] == 0x0D) && (buffer[idx] == 0x0A)) {
                           found = true;
                       } else {
                           idx++;
                       }
                   } else {
                       throw new Exception ("Port closed");
                   }
               }
            } // while !found and !timeout when hunting CRLF
        } // hunt for CR+LF

        if ( ! timer.timedOut()) {
            ret.data = buffer;
            ret.size = idx - 4; // exclude CR, LF and LRC (2bytes)
            ret.success = true;
        } else {
            ret.data = null;
            ret.size = 0;
            ret.success = false;
            ret.exceptionCode = TIMEOUT_ERROR; // timeout
        }
        return ret;
    } // getResponse

    private void unassemblyResponse(CommReturn ret, FunctionCode fc) {
       switch (fc) {
           case ReadRegister:
               unassemblyReadRegister(ret);
               break;
           case WriteRegister:
               // TO_DO: escrever unassemblyWriteRegister(ret);
               break;
           case WriteFile:
               // TO_DO: escrever unassemblyWriteFile(ret);
               break;
       }
    } // unassemblyResponse

    private void unassemblyReadRegister(CommReturn ret) {
        byte fc = decode(ret.data[3], ret.data[4]);
        if ( fc == 3 ) {
            ret.success = true;
            ret.value = decode (ret.data[7], ret.data[8]);
        } else {
            ret.success = false;
            ret.exceptionCode = decode (ret.data[5], ret.data[6]);
        }
    } // unassemblyReadRegister

    private boolean checkLRC(CommReturn ret) {
      int calculatedLRC = calculateLRC(ret.data, 1, ret.size);
      int rxLRC = decode(ret.data[ret.size+1], ret.data[ret.size+2]);
      System.out.println("calculate=" + Integer.toHexString(calculatedLRC) + " rx=" + Integer.toHexString(rxLRC));
      if ( calculatedLRC == rxLRC)
          return true;
      return false;
    } // checkLRC

    private byte[] buildReadRegisterFrame(int destination, Register register) {
        byte[] frame = new byte[15];
        int function03 = 3;
        int startingAddress = addressMap.get(register);

        frame[0] = ':';
        frame[1] = highEncoded(destination);
        frame[2] = lowEncoded(destination);
        frame[3] = highEncoded(function03);
        frame[4] = lowEncoded(function03);
        frame[5] = '0';  // starting address high
        frame[6] = '0';
        frame[7] = highEncoded(startingAddress);
        frame[8] = lowEncoded(startingAddress);
        frame[9] = '0';
        frame[10] = '1';
        byte lrc = calculateLRC(frame, 1, 10);
        frame[11] = highEncoded(lrc);
        frame[12] = lowEncoded (lrc);
        frame[13] = 0x0d;
        frame[14] = 0x0a;
        return frame;
    } // buildReadRegisterFrame

    public int calculateLRC (String txt) {
        byte ret ;
        ret = calculateLRC(txt.getBytes(), 0, txt.length()-1);
        return ret;
    } // calculateLRC for debug

    private byte calculateLRC (byte[] data, int start, int end) {
        byte accum = 0;
        byte ff = (byte)255;
        byte um = (byte)1;

        for (int i=start; i <= end; i++) {
            accum += data[i];
        }
        accum = (byte)(ff - accum);
        return (byte)(accum + um);
    } // calculateLRC

    private byte highEncoded(int value) {
        byte x = (byte)((value & 0xF0)>>4); // encode only a byte
        if ( x < 10) {
          return (byte) (0x30 + x);
        } else {
          return (byte) (0x41 + (x-10));
        }
    } // highASCII

    private byte lowEncoded(int value) {
        byte x = (byte) (value & 0x0F);
        if ( x < 10) {
          return (byte) (0x30 + x);
        } else {
          return (byte) (0x41 + (x-10));
        }
    } // lowEncoded

    private byte decode(byte high, byte low) {
       byte x, y;
       if (low < 'A') {
           x = (byte)(low & 0x0f);
       } else {
           x = (byte)((low & 0x0f) + 0x09);
       }
       if ( high < 'A') {
           y = (byte)(high & 0x0f);
       } else {
           y = (byte)((high & 0x0f) + 0x09);
       }
       return (byte)( x | (y << 4));
    } // decode

    private void connect ( String portName ) throws Exception {
        CommPortIdentifier portIdentifier = CommPortIdentifier.getPortIdentifier(portName);
        if ( portIdentifier.isCurrentlyOwned() ) {
            System.out.println("Error: Port is currently in use");
        }  else {
            CommPort commPort = portIdentifier.open(this.getClass().getName(),2000);
            if ( commPort instanceof SerialPort ) {
                serialPort = (SerialPort) commPort;
                serialPort.setSerialPortParams(9600,SerialPort.DATABITS_8,SerialPort.STOPBITS_1,SerialPort.PARITY_NONE);
                inputSerialStream = serialPort.getInputStream();
                outputSerialStream = serialPort.getOutputStream();
           } else {
                System.out.println("Error: Only serial ports are handled.");
            }
        }
    } // connect

    private void disconnect() {
        try {
            serialPort.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    } // disconnect

} // CommunicationController
